use serde::{Deserialize, Serialize};

use std::fs;

use crate::config;
use crate::time::{unix_to_human_date, unix_to_human_date_time, unix_to_human_time};

/// Structs for handling the Beanconqueror database "Beanconqueror.json" export data

#[derive(Serialize, Deserialize, Debug, Default)]
pub struct Database {
    #[serde(rename = "BEANS")]
    pub beans: Vec<Bean>,

    #[serde(rename = "BREWS")]
    pub brews: Vec<Brew>,

    #[serde(rename = "MILL")]
    pub mill: Vec<Mill>,

    #[serde(rename = "PREPARATION")]
    pub preparation: Vec<Preparation>,
}

#[derive(Serialize, Deserialize, Debug, Default)]
pub struct Bean {
    pub name: String,
    pub config: Config,
}

#[derive(Serialize, Deserialize, Debug, Default)]
pub struct Brew {
    pub bean: String,
    pub config: Config,
    pub flow_profile: Option<String>,
}

#[derive(Serialize, Deserialize, Debug, Default)]
pub struct Mill {
    pub name: String,
    pub config: Config,
}

#[derive(Serialize, Deserialize, Debug, Default)]
pub struct Preparation {
    pub name: String,
    pub config: Config,
    pub tools: Vec<Tools>,
}

#[derive(Serialize, Deserialize, Debug, Default)]
pub struct Tools {
    pub name: String,
    pub config: Config,
}

#[derive(Serialize, Deserialize, Debug, Default)]
pub struct Config {
    pub uuid: String,
    pub unix_timestamp: i64,
}

impl Database {
    pub fn from_file(path: &str) -> Self {
        let file = fs::read_to_string(path)
            .unwrap_or_else(|_| panic!("Cannot open file at path \"{}\"", path));
        Self::from_str(&file)
    }

    pub fn from_str(str: &str) -> Self {
        let database: Self = serde_json::from_str(&str)
            .unwrap_or_else(|_| panic!("Cannot deserialize file from data"));

        database
    }

    pub fn from_config(config: &config::Config) -> Self {
        if config.data.is_some() {
            Database::from_str(&config.get_main_file())
        } else {
            Database::from_file(&config.get_filename(&config.main_json))
        }
    }

    pub fn bean_names(&self) -> Vec<String> {
        self.beans
            .iter()
            .map(|bean| bean.name_with_date())
            .collect()
    }

    pub fn bean_names_with_uuids(&self) -> Vec<(String, String)> {
        self.beans
            .iter()
            .map(|bean| (bean.name_with_date(), bean.config.uuid.to_owned()))
            .collect()
    }

    pub fn bean_for_uuid(&self, uuid: &str) -> Option<&Bean> {
        self.beans
            .iter()
            .find(|bean| bean.config.uuid == uuid.to_owned())
    }

    pub fn brew_for_uuid(&self, uuid: &str) -> Option<&Brew> {
        self.brews
            .iter()
            .find(|brew| brew.config.uuid == uuid.to_owned())
    }

    pub fn brews_for_uuids(&self, uuids: &Vec<String>) -> Vec<&Brew> {
        self.brews
            .iter()
            .filter_map(|brew| match uuids.contains(&brew.config.uuid) {
                true => Some(brew),
                false => None,
            })
            .collect()
    }

    pub fn bean_for_brew(&self, brew: &Brew) -> Option<&Bean> {
        self.beans
            .iter()
            .find(|bean| bean.config.uuid == brew.config.uuid)
    }

    pub fn brews_for_bean(&self, bean: &Bean) -> Vec<&Brew> {
        self.brews
            .iter()
            .filter(|brew| brew.bean == bean.config.uuid)
            .collect()
    }

    // TODO: Preparation, Tools, maybe transiton to HashMaps?
}

impl Bean {
    pub fn name_with_date(&self) -> String {
        format!(
            "{} from {}",
            self.name.to_owned(),
            unix_to_human_date(self.config.unix_timestamp.to_owned())
        )
    }
}

impl Brew {
    pub fn time(&self) -> String {
        unix_to_human_time(self.config.unix_timestamp.to_owned())
    }

    pub fn date_time(&self, database: &Database) -> String {
        if database.bean_for_uuid(&self.bean).is_some() {
            unix_to_human_date_time(self.config.unix_timestamp.to_owned())
        } else {
            String::default()
        }
    }

    pub fn date_time_with_bean(&self, database: &Database) -> String {
        if let Some(bean) = &database.bean_for_uuid(&self.bean) {
            format!(
                "{} ({})",
                unix_to_human_date_time(self.config.unix_timestamp.to_owned()),
                bean.name_with_date()
            )
        } else {
            String::default()
        }
    }
}
