use crate::time::*;

use chrono::NaiveTime;

use serde::{Deserialize, Serialize};

use std::fs;

/// Structs for handling the Beanconqueror "*_flow_profile.json" export data

#[derive(Serialize, Deserialize, Debug)]
pub struct FlowProfile {
    pub weight: Vec<WeightData>,
    #[serde(rename = "waterFlow")]
    pub water_flow: Vec<FlowData>,
    #[serde(rename = "realtimeFlow")]
    pub realtime_flow: Vec<RtFlowData>,

    #[serde(skip)]
    pub cutoff: Option<f64>,
    #[serde(skip)]
    pub data_collection: Option<DataCollection>,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct WeightData {
    pub timestamp: String,
    pub brew_time: String,
    pub actual_weight: f64,
    pub old_weight: f64,
    pub actual_smoothed_weight: f64,
    pub old_smoothed_weight: f64,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct FlowData {
    pub brew_time: String,
    pub timestamp: String,
    pub value: f64,
}

#[derive(Serialize, Deserialize, Debug)]
pub struct RtFlowData {
    pub brew_time: String,
    pub timestamp: String,
    pub smoothed_weight: f64,
    pub flow_value: f64,
}

#[derive(Default, Debug)]
pub struct DataCollection {
    pub reference_time: NaiveTime,
    pub weight: Vec<(f64, f64)>,
    pub flow: Vec<(f64, f64)>,
    pub rt_flow: Vec<(f64, f64)>,
}

impl FlowProfile {
    pub fn from_file(path: &str, cutoff: Option<f64>) -> Self {
        let file = fs::read_to_string(path)
            .unwrap_or_else(|_| panic!("Cannot open file at path \"{}\"", path));
        FlowProfile::from_str(&file, cutoff)
    }

    pub fn from_str(str: &str, cutoff: Option<f64>) -> Self {
        let mut brew: Self =
            serde_json::from_str(&str).unwrap_or_else(|_| panic!("Cannot deserialize file"));
        brew.cutoff = cutoff;

        brew
    }

    pub fn preprocess_json_mut(&mut self) {
        self.data_collection = self.process_json();
    }

    pub fn preprocess_json(self) -> Self {
        let data_collection = self.process_json();
        Self {
            weight: self.weight,
            water_flow: self.water_flow,
            realtime_flow: self.realtime_flow,
            cutoff: None,
            data_collection,
        }
    }

    pub fn process_json(&self) -> Option<DataCollection> {
        let weight_starting_time: NaiveTime = str_to_naivetime(&(self.weight[0].timestamp));
        let flow_starting_time: NaiveTime = str_to_naivetime(&(self.water_flow[0].timestamp));
        let rt_flow_starting_time: NaiveTime = str_to_naivetime(&(self.realtime_flow[0].timestamp));

        let reference_time: NaiveTime = weight_starting_time
            .min(flow_starting_time)
            .min(rt_flow_starting_time);

        let mut weight_tuples: Vec<(f64, f64)> = self
            .weight
            .iter()
            .map(
                |WeightData {
                     ref timestamp,
                     ref actual_weight,
                     ..
                 }| {
                    let deltatime = deltatime(str_to_naivetime(&timestamp), reference_time);
                    let std_duration = deltatime.to_std().unwrap();

                    (std_duration.as_secs_f64(), actual_weight.to_owned())
                },
            )
            .collect();

        let mut flow_tuples: Vec<(f64, f64)> = self
            .water_flow
            .iter()
            .map(
                |FlowData {
                     ref timestamp,
                     ref value,
                     ..
                 }| {
                    let deltatime = deltatime(str_to_naivetime(&timestamp), reference_time);
                    let std_duration = deltatime.to_std().unwrap();

                    (std_duration.as_secs_f64(), value.to_owned())
                },
            )
            .collect();

        let mut rt_flow_tuples: Vec<(f64, f64)> = self
            .realtime_flow
            .iter()
            .map(
                |RtFlowData {
                     ref timestamp,
                     ref flow_value,
                     ..
                 }| {
                    let deltatime = deltatime(str_to_naivetime(&timestamp), reference_time);
                    let std_duration = deltatime.to_std().unwrap();

                    (std_duration.as_secs_f64(), flow_value.to_owned())
                },
            )
            .collect();

        if let Some(cutoff_val) = self.cutoff {
            if cutoff_val != -1.0 {
                weight_tuples.retain(|tuple| tuple.0 < cutoff_val);
                flow_tuples.retain(|tuple| tuple.0 < cutoff_val);
                rt_flow_tuples.retain(|tuple| tuple.0 < cutoff_val);
            }
        }

        if !weight_tuples.is_empty() && !flow_tuples.is_empty() && !rt_flow_tuples.is_empty() {
            Some(DataCollection {
                reference_time,
                weight: weight_tuples,
                flow: flow_tuples,
                rt_flow: rt_flow_tuples,
            })
        } else {
            None
        }
    }
}
