use crate::config::Config;
use crate::database::{Brew, Database};
use crate::flow_profile::FlowProfile;

use egui::plot::{PlotPoint, PlotPoints};

use std::collections::HashMap;
use std::sync::{Arc, Mutex};

pub type RawPlotPoints = Vec<PlotPoint>;
pub type PlotEntry = (String, RawPlotPoints);

#[derive(Default)]
pub struct LoadingProgress {
    pub curr_name: String,
    pub current: usize,
    pub total: usize,
    pub percentage: f32,
    pub finished: bool,
}

pub fn plot_points_to_owned(plot_points: &RawPlotPoints) -> PlotPoints {
    PlotPoints::Owned(plot_points.to_owned())
}

pub fn database_plot_entry(brew: &Brew, config: &Config, database: &Database) -> Option<PlotEntry> {
    let mut plot_points: Option<(String, RawPlotPoints)> = None;
    if let Some(brew_flow_profile) = &brew.flow_profile {
        if !&brew_flow_profile.is_empty() {
            let brew_title = brew.date_time_with_bean(&database);

            let flow_profile: FlowProfile;
            let flow_profile_filename = &config.get_filename(&brew_flow_profile);
            if let Some(file) = &config.get_file(&flow_profile_filename) {
                flow_profile = FlowProfile::from_str(&file, None).preprocess_json();
            } else {
                flow_profile =
                    FlowProfile::from_file(&flow_profile_filename, None).preprocess_json();
            }

            if let Some(data_collection) = &flow_profile.data_collection {
                let point_vec: Vec<[f64; 2]> = data_collection
                    .weight
                    .iter()
                    .map(|(x, y)| [x.to_owned(), y.to_owned()])
                    .collect();

                plot_points = Some((brew_title, PlotPoints::from(point_vec).points().to_vec()));
            }
        }
    }

    plot_points
}

pub fn database_plot_entries(
    brews: Vec<&Brew>,
    config: &Config,
    database: &Database,
    progress: Option<Arc<Mutex<LoadingProgress>>>,
) -> HashMap<String, PlotEntry> {
    let mut plot_entries: HashMap<String, PlotEntry> = HashMap::new();
    let step_size = 1.0 / brews.len() as f32;

    if let Some(progress) = &progress {
        let mut progress_lock = progress.lock().unwrap();
        (*progress_lock).total = brews.len();
        (*progress_lock).finished = false;
    }

    for (i, brew) in brews.iter().enumerate() {
        if let Some(progress) = &progress {
            let mut progress_lock = progress.lock().unwrap();
            (*progress_lock).curr_name = brew.date_time_with_bean(&database);
            (*progress_lock).current = i + 1;
            (*progress_lock).percentage += step_size;
        }

        if let Some(plot_entry) = database_plot_entry(&brew, &config, &database) {
            plot_entries.insert(brew.config.uuid.to_owned(), plot_entry);
        }
    }

    if let Some(progress) = &progress {
        let mut progress_lock = progress.lock().unwrap();
        (*progress_lock).finished = true;
    }

    plot_entries
}
