use iced::alignment::{self, Alignment};
use iced::widget::{Button, Checkbox};
use iced::widget::{button, checkbox, column, container, row, scrollable, text};
use iced::Length;
use iced::{Element, Sandbox};

use crate::config::Config;
use crate::database::Database;
use crate::plot::database_plot_selected;

use std::process::Command;

#[derive(Debug)]
pub enum Selector {
    Loading,
    Loaded(State),
}

#[derive(Debug)]
pub struct State {
    config: Config,
    database: Database,
    items: Vec<Item>,
}

#[derive(Debug, Clone)]
pub enum Message {
    Placeholder,
    GenerateClicked,
    OpenClicked,
    ItemMessage(usize, ItemMessage),
}

impl Sandbox for Selector {
    type Message = Message;

    fn new() -> Self {
        let mut selector = Selector::Loading;
        selector.update(Message::Placeholder);

        selector
    }

    fn title(&self) -> String {
        String::from("RustyBeans")
    }

    fn update(&mut self, message: Message) {
        match self {
            Selector::Loading => {
                let config = Config::from_default();
                let database = Database::from_config(&config);
                let items = database
                    .bean_names_with_uuids()
                    .into_iter()
                    .map(|(name, uuid)| Item::new(name, uuid))
                    .collect();

                *self = Selector::Loaded(State {
                    config,
                    database,
                    items,
                });
            }

            Selector::Loaded(state) => match message {
                Message::GenerateClicked => {
                    database_plot_selected(
                        state
                            .items
                            .iter()
                            .filter(|item| item.checked)
                            .map(|item| item.uuid.to_owned())
                            .collect(),
                    );
                }

                Message::OpenClicked => {
                    #[cfg(target_os = "linux")]
                    Command::new("xdg-open")
                        .arg(&state.config.output_dir)
                        .spawn()
                        .unwrap();

                    #[cfg(target_os = "windows")]
                    Command::new("explorer")
                        .arg(&state.config.output_dir)
                        .spawn()
                        .unwrap();

                    #[cfg(target_os = "macos")]
                    Command::new("open")
                        .arg(&state.config.output_dir)
                        .spawn()
                        .unwrap();
                }

                Message::ItemMessage(i, item_message) => {
                    if let Some(item) = state.items.get_mut(i) {
                        item.update(item_message);
                    }
                }

                _ => {}
            },
        }
    }

    fn view(&self) -> Element<Message> {
        match self {
            Selector::Loading => container(
                text("Loading data ...")
                    .horizontal_alignment(alignment::Horizontal::Center)
                    .size(50),
            )
            .width(Length::Fill)
            .height(Length::Fill)
            .center_x()
            .center_y()
            .into(),

            Selector::Loaded(State { items, .. }) => {
                let title = text("RustyBeans")
                    .width(Length::Fill)
                    .size(50)
                    .horizontal_alignment(alignment::Horizontal::Center);

                let items: Element<_> = column(
                    items
                        .iter()
                        .enumerate()
                        .map(|(i, item)| {
                            item.view()
                                .map(move |message| Message::ItemMessage(i, message))
                        })
                        .collect(),
                )
                .spacing(10)
                .into();

                let generate_btn: Button<_> = button(text("Generate Charts").size(24))
                    .on_press(Message::GenerateClicked)
                    .padding(8);
                let open_btn: Button<_> = button(text("Open Folder").size(24))
                    .on_press(Message::OpenClicked)
                    .padding(8);

                let button_row = row![generate_btn, open_btn].spacing(20);
                let content = column![title, items, button_row].spacing(20).max_width(800);

                scrollable(
                    container(content)
                        .width(Length::Fill)
                        .padding(40)
                        .center_x(),
                )
                .into()
            }
        }
    }
}

#[derive(Debug, Clone)]
struct Item {
    title: String,
    uuid: String,
    checked: bool,
}

#[derive(Debug, Clone)]
pub enum ItemMessage {
    Checked(bool),
}

impl Item {
    fn new(title: String, uuid: String) -> Self {
        Self {
            title,
            uuid,
            checked: false,
        }
    }

    fn update(&mut self, message: ItemMessage) {
        match message {
            ItemMessage::Checked(checked) => {
                self.checked = checked;
            }
        }
    }

    fn view(&self) -> Element<ItemMessage> {
        let checkbox: Checkbox<_> = checkbox(&self.title, self.checked, |checked| {
            ItemMessage::Checked(checked)
        })
        .width(Length::Fill);

        row![checkbox,]
            .spacing(20)
            .align_items(Alignment::Center)
            .into()
    }
}
